
import java.awt.Color;
import java.awt.Graphics;
import java.awt.FontMetrics;

class DrawableWord
	{
	/*
	Implement a moveable, self-drawing word set inside a colored rectangle.
	*/

	//a word has an (x, y) position, a width and height, and a name
	private int x, y;
	private int width, height;
	private String name;

	//it may also be currently selected or not
	private boolean selected = false;

	//amount of horizontal and vertical padding around the word (in pixels)
	public final static int X_INSET = 5;
	public final static int Y_INSET = 5;

	//all words have a normal color,
	private final static Color color = new Color(250, 220, 100);
	private final static Color darkColor = gain(color, 0.71);
	private final static Color lightColor = gain(color, 1.31);

	//and a color to change to when they're selected
	private final static Color selectedColor = Color.pink;
	private final static Color selectedDark = gain(selectedColor, 0.71);
	private final static Color selectedLight = gain(selectedColor, 1.31);

	////////////////////////////////////////////////////////////////////////
	public DrawableWord(String name)
		{
		this.name = name;
		}

	////////////////////////////////////////////////////////////////////////
	public int getXPosition()
		{
		return x;
		}

	////////////////////////////////////////////////////////////////////////
	public int getYPosition()
		{
		return y;
		}

	////////////////////////////////////////////////////////////////////////
	public String getName()
		{
		return name;
		}

	////////////////////////////////////////////////////////////////////////
	public int getWidth()
		{
		return width;
		}

	////////////////////////////////////////////////////////////////////////
	public int getHeight()
		{
		return height;
		}

	////////////////////////////////////////////////////////////////////////
	public void setWidth(int width)
		{
		this.width = width;
		}

	////////////////////////////////////////////////////////////////////////
	public void setHeight(int height)
		{
		this.height = height;
		}

	////////////////////////////////////////////////////////////////////////
	public void setSelected(boolean setting)
		{
		selected = setting;
		}

	////////////////////////////////////////////////////////////////////////
	public void moveTo(int x, int y)
		{
		/*
		Move this DrawableWord to the position (x, y).
		*/

		this.x = x; this.y = y;
		}

	////////////////////////////////////////////////////////////////////////
	public boolean contains(int x, int y)
		{
		/*
		Is the point (x, y) somewhere within this DrawableWord?
		*/

		boolean xIntersected = ((x >= this.x) && (x < this.x + width));
		boolean yIntersected = ((y >= this.y) && (y < this.y + height));

		return (xIntersected && yIntersected);
		}

	////////////////////////////////////////////////////////////////////////
	public boolean intersects(DrawableWord word)
		{
		/*
		Does this DrawableWord intersect the given DrawableWord?
		*/

		boolean xRangesIntersect =
			(word.getXPosition() < x + width) &&
			(word.getXPosition() + word.getWidth() > x);

		boolean yRangesIntersect =
			(word.getYPosition() < y + height) &&
			(word.getYPosition() + word.getHeight() > y);

		return (xRangesIntersect && yRangesIntersect);
		}

	////////////////////////////////////////////////////////////////////////
	public void adjustPosition(DrawableWord word, int xOrigin, int yOrigin)
		{
		/*
		Position this DrawableWord relative to the origin of a given
		DrawableWord.
		*/

		x += word.getXPosition() - xOrigin;
		y += word.getYPosition() - yOrigin;
		}
	
	///////////////////////////////////////////////////////////////////////
	public void paint(Graphics graphicsContext, FontMetrics fontMetrics)
		{
		/*
		Paint the rectangle and the name.
		*/

		//use a different color if it's currently selected
		if (selected)
			paint(graphicsContext, selectedColor, selectedLight, selectedDark);
		else
			paint(graphicsContext, color, lightColor, darkColor);

		//write the name into the rectangle
		graphicsContext.setColor(Color.black);
		graphicsContext.drawString(name, x + X_INSET,
								y + Y_INSET + fontMetrics.getAscent());
		}

	////////////////////////////////////////////////////////////////////////
	private void paint(Graphics graphicsContext, Color color,
							Color lightColor, Color darkColor)
		{
		/*
		Paint the rectangle. Make it look 3d if it's currently selected.
		*/

		//color the whole rectangle
		graphicsContext.setColor(color);
		graphicsContext.fillRect(x, y, width, height);

		if (selected)
			graphicsContext.setColor(darkColor);
		else
			graphicsContext.setColor(lightColor);
		graphicsContext.fillRect(x, y, width - 1, 1);
		graphicsContext.fillRect(x, y + 1, 1, height - 2);

		if (selected)
			graphicsContext.setColor(lightColor);
		else
			graphicsContext.setColor(Color.black);
		graphicsContext.fillRect(x, y + height - 1, width, 1);
		graphicsContext.fillRect(x + width - 1, y, 1, height - 1);

		if (selected)
			{
			graphicsContext.setColor(Color.black);
			graphicsContext.fillRect(x + 1, y + 1, width - 3, 1);
			graphicsContext.fillRect(x + 1, y + 2, 1, height - 4);
			}
		else
			{
			graphicsContext.setColor(darkColor);
			graphicsContext.fillRect(x + 1, y + height - 2, width - 2, 1);
			graphicsContext.fillRect(x + width - 2, y + 1, 1, height - 3);
			}
		}

	////////////////////////////////////////////////////////////////////////
	public String toString()
		{
		return "DrawableWord['" + name + "' (" + x + "," + y + ") " +
					width + "x" + height + "]";
		}

	///////////////////////////////////////////////////////////////////////
	private static Color gain(Color color, double gain)
		{
		/*
		Generate a new color with higher (or lower) gain than the given color.
		*/

		int red = clamp(color.getRed() * gain);
		int green = clamp(color.getGreen() * gain);
		int blue = clamp(color.getBlue() * gain);

		return new Color(red, green, blue);
		}

	///////////////////////////////////////////////////////////////////////
	private static int clamp(double colorValue)
		{
		/*
		Make sure the given colorValue stays in the range [0, 255].
		*/

		if (colorValue < 0)
			colorValue = 0;
		if (colorValue > 255)
			colorValue = 255;

		return (int) colorValue;
		}
	}
