
import DrawableWord;

import java.awt.Image;
import java.awt.Graphics;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Panel;
import java.awt.Event;
import java.util.Vector;
import java.util.Enumeration;

class WordPanel extends Panel
	{
	//the array of words
	private DrawableWord[] words;
	private int numWords;

	//whether the words are to be pushable
	private boolean wordsArePushable;

	//number of pixels to indent a word within its rectangle
	private final static int X_INSET = DrawableWord.X_INSET;
	private final static int Y_INSET = DrawableWord.Y_INSET;

	//number of times to try to reposition words to avoid word overlaps
	//during scrambling
	private final static int NUM_TRIES = 30;

	private DrawableWord currentSelectedWord;
	private DrawableWord lastSelectedWord;
	private DrawableWord currentOverlappedWord;
	private DrawableWord firstOverlappedWord;
	private DrawableWord lastOverlappedWord;

	//to remember the relative position of the mouse and any selected word
	private int xOffset, yOffset;

	//the two offscreen bitmaps, their graphics contexts,
	//and their (common) dimensions and font metrics
	private Image offscreenImage1, offscreenImage2;
	private Graphics offscreenGraphics1, offscreenGraphics2;
	private Dimension offscreenImageSize;
	private FontMetrics fontMetrics;

	//this panel's window dimensions
	private int windowWidth, windowHeight;

	///////////////////////////////////////////////////////////////////////
	public WordPanel(Vector names)
		{
		/*
		Initialize a WordPanel containing a collection of words stored in
		the given Vector.
		*/

		//initialize the offscreen images, their graphics contexts, and fonts
		checkIfResized();

		//create the DrawableWord objects
		numWords = names.size();
		words = new DrawableWord[numWords];
		for (int i = 0; i < numWords; i++)
			words[i] = new DrawableWord((String) names.elementAt(i));

		wordsArePushable = false;

		//scatter the words randomly over the panel's window
		scramble();
		}

	///////////////////////////////////////////////////////////////////////
	public void scramble()
		{
		/*
		Randomize the positions of all the words.
		*/

		for (int i = 0; i < numWords; i++)
			{
			boolean wordsOverlap;
			int iterations = 0;
			do
				{
				wordsOverlap = false;
				words[i].moveTo((int) (X_INSET + (windowWidth - 2 * X_INSET -
									words[i].getWidth()) * Math.random()),
									(int) (Y_INSET + (windowHeight - 2 * Y_INSET -
									words[i].getHeight()) * Math.random()));

				if (iterations < NUM_TRIES)
					for (int j = 0; j < i; j++)
						if (words[i].intersects(words[j]))
							wordsOverlap = true;
				iterations++;
				}
			while(wordsOverlap);
			}
		}

	///////////////////////////////////////////////////////////////////////
	public void setWordsPushable(boolean setting)
		{
		wordsArePushable = setting;
		}

	///////////////////////////////////////////////////////////////////////
	public boolean mouseDown(Event event, int mouseXPosition, int mouseYPosition)
		{
		/*
		Check whether the user has clicked on a word; if so, select it.

		If a word is to be selected, remember the relative location of the mouse
		within the word's rectangle with the vector (xOffset, yOffset).
		*/

		int wordIndex = indexOfWordContaining(mouseXPosition, mouseYPosition);
		if (wordIndex >= 0)
			{
			moveWordToEnd(wordIndex);
			currentSelectedWord = words[numWords - 1];
			firstOverlappedWord = firstIntersectingWord(currentSelectedWord);
			xOffset = currentSelectedWord.getXPosition() - mouseXPosition;
			yOffset = currentSelectedWord.getYPosition() - mouseYPosition;
			repaint();
			}

		return true;
		}

	///////////////////////////////////////////////////////////////////////
	public boolean mouseDrag(Event event, int mouseXPosition, int mouseYPosition)
		{
		/*
		Move the currently selected word, if any.
		If words are currently pushable, move the first overlapped word, if any.
		*/

		if (currentSelectedWord != null)
			{
			int xOrigin = currentSelectedWord.getXPosition();
			int yOrigin = currentSelectedWord.getYPosition();
			currentSelectedWord.moveTo(mouseXPosition + xOffset,
												mouseYPosition + yOffset);
			currentOverlappedWord = firstIntersectingWord(currentSelectedWord);
			if (currentOverlappedWord == null)
				firstOverlappedWord = null;
			else
				if (wordsArePushable &&
					(firstOverlappedWord != currentOverlappedWord))
					currentOverlappedWord.adjustPosition(currentSelectedWord, xOrigin, yOrigin);
			repaint();
			}

		return true;
		}

	///////////////////////////////////////////////////////////////////////
	public boolean mouseUp(Event event, int mouseXPosition, int mouseYPosition)
		{
		/*
		The mouse has been released; if there is a currently selected word,
		deselect it.
		*/

		if (currentSelectedWord != null)
			{
			currentSelectedWord.moveTo(mouseXPosition + xOffset,
												mouseYPosition + yOffset);
			currentSelectedWord = null;
			currentOverlappedWord = null;
			repaint();
			}

		return true;
		}

	///////////////////////////////////////////////////////////////////////
	public void update(Graphics graphicsContext)
		{
		/*
		Override Panel.update() to reduce flicker.
		*/

		paint(graphicsContext);
		}

	///////////////////////////////////////////////////////////////////////
	public void paint(Graphics graphicsContext)
		{
		/*
		Paint the entire panel of words.
		*/

		checkIfResized();

		if ((currentSelectedWord == null) ||
			(lastSelectedWord != currentSelectedWord) ||
			(wordsArePushable && (lastOverlappedWord != currentOverlappedWord)))
			{
			//clear the first offscreen bitmap
			offscreenGraphics1.setColor(this.getBackground());
			offscreenGraphics1.fillRect(0, 0, windowWidth, windowHeight);

			//paint all non-selected, non-overlapped words into it
			for (int i = 0; i < numWords; i++)
				if ((words[i] != currentSelectedWord) &&
					(words[i] != currentOverlappedWord))
					paintWord(offscreenGraphics1, words[i], fontMetrics);

			//remember the selected and overlapped words (if any) for next time
			lastSelectedWord = currentSelectedWord;
			lastOverlappedWord = currentOverlappedWord;
			}

		//draw the whole mess to the second offscreen bitmap
		offscreenGraphics2.drawImage(offscreenImage1, 0, 0, null);

		//draw the selected and overlapped words (if any) into the second bitmap
		//(so they will be on top of any of the words already there)
		if (wordsArePushable && (currentOverlappedWord != null))
			paintWord(offscreenGraphics2, currentOverlappedWord, fontMetrics);
		if (currentSelectedWord != null)
			paintWord(offscreenGraphics2, currentSelectedWord, fontMetrics);

		//draw the second bitmap to the screen
		graphicsContext.drawImage(offscreenImage2, 0, 0, null);
		}

	///////////////////////////////////////////////////////////////////////
	private void paintWord(Graphics graphicsContext, DrawableWord word,
								FontMetrics fontMetrics)
		{
		/*
		Draw a word into the given graphics context.
		*/

		if (word == currentSelectedWord)
			word.setSelected(true);
		else
			word.setSelected(false);
		word.paint(graphicsContext, fontMetrics);
		}

	///////////////////////////////////////////////////////////////////////
	private int indexOfWordContaining(int x, int y)
		{
		/*
		Check whether any of the words contains the given (x, y) position.
		Return the index of last containing word, if any---counting from
		the beginning of the word list.
		*/

		for (int i = numWords - 1; i >= 0; i--)
			if (words[i].contains(x, y))
				return i;

		return -1;
		}

	///////////////////////////////////////////////////////////////////////
	private DrawableWord firstIntersectingWord(DrawableWord word)
		{
		/*
		Check whether any of the words intersects the given word.
		Return the first intersecting word, if any---counting from the
		beginning of the word list.
		*/

		for (int i = 0; i < numWords - 1; i++)
			if (word.intersects(words[i]))
				return words[i];

		return null;
		}
	
	///////////////////////////////////////////////////////////////////////
	private void moveWordToEnd(int index)
		{
		/*
		Move the indexth word to the end of the list of words,
		thereby making it appear to be on top of all the other words
		(paint() draws the words from the beginning of the list to the end).
		*/

		DrawableWord word = words[index];
		for (int i = index; i < numWords - 1; i++)
			words[i] = words[i + 1];
		words[numWords - 1] = word;
		}

	///////////////////////////////////////////////////////////////////////
	private void checkIfResized()
		{
		/*
		Check whether this is the first time the screen has been created,
		or if the window has been resized. If so, (re)create the offscreen
		bitmaps and resize the words.
		*/

		//save the current panel dimensions
		windowWidth = this.getSize().width;
		windowHeight = this.getSize().height;

		if ((windowWidth < 1) || (windowHeight < 1))
			return;

		if ((offscreenImage1 == null) ||
			(offscreenImageSize.width != windowWidth) ||
			(offscreenImageSize.height != windowHeight))
			{
			//create the two offscreen bitmaps and get their graphics contexts
			offscreenImage1 = createImage(windowWidth, windowHeight);
			offscreenImage2 = createImage(windowWidth, windowHeight);
			offscreenGraphics1 = offscreenImage1.getGraphics();
			offscreenGraphics2 = offscreenImage2.getGraphics();

			//set their (common) font, font metrics, and dimension
			offscreenGraphics1.setFont(this.getFont());
			offscreenGraphics2.setFont(this.getFont());
			fontMetrics = offscreenGraphics1.getFontMetrics();
			offscreenImageSize = this.getSize();

			//resize the words
			for (int i = 0; i < numWords; i++)
				{
				words[i].setHeight(fontMetrics.getHeight() + 2 * Y_INSET);
				words[i].setWidth(fontMetrics.stringWidth(words[i].getName()
										+ 2 * X_INSET));
				}

			//scatter the words randomly over the panel
			scramble();
			}
		}
	}
